import click
from stairs import get_project


@click.group()
def consumer_cli():
    pass


@consumer_cli.command("consumer:standalone")
@click.argument('name')
@click.option('--noprint', '-np', is_flag=True, help="Disable print")
def init_session(name, noprint):
    """
    Start listening pipelines and then execute jobs which were sent to specified
    consumer.
    """
    project = get_project()
    project.set_verbose(not noprint)

    if project.verbose:
        print("Standalone consumer started, and listening for a tasks ...")
        print("Press CTRL-C to stop listening")

    consumer = get_project().get_consumer_by_name(name)

    if consumer is None:
        raise RuntimeError("Consumer not found")

    consumer.run_worker()




from stepist.flow.steps.step import StepData

import ujson


def get_connection(sqs_engine):
    queues = sqs_engine._queues.keys()
    return SQSConnection(sqs_engine.session,
                         queues,
                         sqs_engine.message_retention_period,
                         sqs_engine.visibility_timeout)


class SQSConnection:
    def __init__(self, sqs_session, queues_keys, message_retention_period,
                 visibility_timeout):
        self.sqs_session = sqs_session
        self.queues_keys = queues_keys

        self.message_retention_period = message_retention_period
        self.visibility_timeout = visibility_timeout

        self.connection = None
        self.channel = None
        self.queues = None

    def init_connection(self):
        self.sqs_client = self.sqs_session.client('sqs')
        self.sqs_resource = self.sqs_session.resource('sqs')

        for q in self.queues_keys:
            self.register_worker(q)

    def register_worker(self, queue_name):
        attrs = {}
        kwargs = {
            'QueueName': queue_name,
            'Attributes': attrs,
        }
        if self.message_retention_period is not None:
            attrs['MessageRetentionPeriod'] = str(self.message_retention_period)
        if self.visibility_timeout is not None:
            attrs['VisibilityTimeout'] = str(self.visibility_timeout)

        self.sqs_client.create_queue(**kwargs)

        queue = self.sqs_resource.get_queue_by_name(QueueName=queue_name)

        self.queues[queue_name] = queue

    def add_job(self, step_key: str, data):
        queue = self.queues.get(step_key, None)
        if not queue:
            raise RuntimeError("Queue %s not found" % step_key)

        step_data = StepData(data)
        step_data_str = ujson.dumps({'data': step_data.get_dict()})

        kwargs = {
            'MessageBody': step_data_str,
            'MessageAttributes': {},
            'DelaySeconds': 0
        }

        ret = queue.send_message(**kwargs)
        return ret['MessageId']




import click
from stairs import get_project
from multiprocessing import Process


@click.group()
def workers_cli():
    pass


@workers_cli.command("pipelines:run")
@click.argument("pipelines", nargs=-1, default=None)
@click.option('--processes', '-p', nargs=1, default=1,
              help="Amount of processes to run in parallel")
@click.option('--noprint', '-np', is_flag=True, help="Disable print")
def run(pipelines, noprint, processes):
    """
    Run all or defined pipelines. Process listening for a jobs until you
    press CTRL-C to exit.

    Use `pipelines:run` to run all pipelines.
    Use `pipelines:run pipeline_name` to run specific pipeline.

    """
    project = get_project()
    project.set_verbose(not noprint)

    if project.verbose:
        print("Pipelines started")

    pipelines_to_run = []
    if pipelines is not None:
        for p in pipelines:
            p_to_run = get_project().get_pipeline_by_name(p)
            if p_to_run is None:
                raise RuntimeError("Pipeline `%s` not found" % p)
            pipelines_to_run.append(p_to_run)
    else:
        pipelines_to_run = None

    processes_objects = []
    for i in range(processes):
        p = Process(target=project.run_pipelines,
                    kwargs=dict(custom_pipelines_to_run=pipelines_to_run))
        p.start()
        processes_objects.append(p)

    for p in processes_objects:
        p.join()



import time
import logging

from stairs.core.worker.worker import Pipeline

logger = logging.getLogger(__name__)


def producer_retry(retries_amount, exceptions, sleep_time=1):

    def wrapper(func):
        def wrap_func(*args, **kwargs):

            for i in range(retries_amount):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    logger.warning(e)
                    time.sleep(sleep_time)

            return func(*args, **kwargs)
        return wrap_func
    return wrapper


def custom_callbacks_to_dict(custom_callbacks: list) -> dict:
    callbacks_key_value = dict()
    for c in custom_callbacks:
        name = c.get_handler_name() if isinstance(c, Pipeline) else c.__name__
        callbacks_key_value[name] = c

    return callbacks_key_value


from stepist.flow.workers.adapters import simple_queue, rm_queue, sqs_queue

from stairs.services import spark as spark_workers
from stairs.core.app import components
from stairs.core.producer.utils import custom_callbacks_to_dict


class SparkProducer(components.AppProducer):
    """

    """
    DEFAULT_QUEUE_LIMIT = 10 ** 6

    def __init__(self, app, handler, default_callbacks: list,
                 custom_callbacks: list, queue_limit=None):

        self.app = app

        self.queue_limit = queue_limit or self.DEFAULT_QUEUE_LIMIT

        # The main generator which yields data
        self.handler = handler

        # Callbacks which should be run always
        self.default_callbacks = default_callbacks or []
        # Callbacks which should be run based on user console, input
        self.custom_callbacks = custom_callbacks_to_dict(custom_callbacks or [])

        components.AppProducer.__init__(self, app)

    def __call__(self, *args, **kwargs):
        self.run(user_args=args, user_kwargs=kwargs)

    def run(self, custom_callbacks_keys: list = None,
            user_args=None, user_kwargs=None):
        """
        Execute producer from console with specified args and kwargs.
        Also can have custom callbacks specified there.
        """
        custom_callbacks = []

        # Basic check for custom producers
        for custom_callback in custom_callbacks_keys or []:
            callback = self.custom_callbacks.get(custom_callback, None)
            if callback is None:
                print("Producer callback `%s` (another producer or pipeline"
                      ") not found." % custom_callback)
                exit()
            custom_callbacks.append(callback)

        # Basic check for callbacks
        if not custom_callbacks and not self.default_callbacks:
            print("No callbacks was found, specified default callback or use"
                  "custom callback")
            exit()

        callbacks_to_run = custom_callbacks + self.default_callbacks

        user_args = user_args or []
        user_kwargs = user_kwargs or dict()

        worker_engine = self.app.project.stepist_app.worker_engine

        if isinstance(worker_engine, simple_queue.SimpleQueueAdapter):
            spark_worker = spark_workers\
                .redis_queue\
                .get_connection(worker_engine)
        elif isinstance(worker_engine, rm_queue.RQAdapter):
            spark_worker = spark_workers \
                .rm_queue \
                .get_connection(worker_engine)
        elif isinstance(worker_engine, sqs_queue.SQSAdapter):
            spark_worker = spark_workers \
                .sqs_queue \
                .get_connection(worker_engine)
        else:
            raise RuntimeError("Spark don't support current queue broken")

        steps_keys_to_run = [c.step.step_key() for c in callbacks_to_run]
        # Running jobs from producer
        spark_rdd = self.handler(*user_args, **user_kwargs)

        SparkJobs(spark_worker, steps_keys_to_run).show_must_go_on(spark_rdd)

    def flush(self):
        for pipeline in self.default_callbacks:
            pipeline.step.flush_all()

        for pipeline in self.custom_callbacks.values():
            pipeline.step.flush_all()

    def get_producer_id(self):
        return "producer:%s:%s" % (self.app.app_name, self.handler.__name__)

    def get_handler_name(self):
        return self.handler.__name__

    def key(self):
        return self.get_handler_name()


class SparkJobs:
    def __init__(self, spark_worker, steps_keys):
        self.spark_worker = spark_worker
        self.steps_keys = steps_keys

    def show_must_go_on(self, spark_rdd):
        spark_rdd.foreachPartition(self.handle_rdd)

    def handle_rdd(self, rdd):
        self.spark_worker.init_connection()
        for item in rdd:
            self(item.asDict())

    def __call__(self, row_data):
        for key in self.steps_keys:
            self.spark_worker.add_job(key, row_data)



from contextlib import contextmanager

from . import global_stairs


def storage():
    if not hasattr(global_stairs, 'producer'):
        global_stairs.producer = dict(
            custom_callbacks=None,
        )

    return global_stairs.producer


def set_custom_callbacks(custom_callbacks):
    storage()['custom_callbacks'] = custom_callbacks


def get_custom_callbacks():
    return storage()['custom_callbacks']


@contextmanager
def change_custom_callbacks(custom_callbacks):
    prev_custom_callbacks = get_custom_callbacks()
    try:
        set_custom_callbacks(custom_callbacks)
        yield
    finally:
        set_custom_callbacks(prev_custom_callbacks)



from stairs.core.producer.producer import Producer
from stairs.core.producer.batch_producer import BatchProducer
from stairs.core.producer.spark_producer import SparkProducer
from stairs.core.output.output_model import Output
from stairs.core.output.standalone import StandAloneConsumer
from stairs.core.output.consumer_iter import ConsumerIter
from stairs.core.worker.worker import Pipeline


class ComponentsMixin:
    components = None

    def producer(self, *pipelines, custom: list=None) -> Producer:
        def _handler_wrap(handler) -> Producer:
            custom_callbacks = custom or self.components.pipelines.values()
            producer = Producer(app=self,
                                handler=handler,
                                default_callbacks=list(pipelines),
                                custom_callbacks=custom_callbacks)

            return producer

        return _handler_wrap

    def batch_producer(self, producer: Producer) -> BatchProducer:
        def _handler_wrap(handler):
            batch_producer = BatchProducer(app=self,
                                           handler=handler,
                                           simple_producer=producer)
            return batch_producer
        return _handler_wrap

    def spark_producer(self, *pipelines, custom: list=None) -> SparkProducer:
        def _handler_wrap(handler) -> Producer:
            custom_callbacks = custom or self.components.pipelines.values()
            producer = SparkProducer(app=self,
                                     handler=handler,
                                     default_callbacks=list(pipelines),
                                     custom_callbacks=custom_callbacks)

            return producer

        return _handler_wrap

    def pipeline(self, config=None) -> Pipeline:
        # TODO: add custom queue name for pipelines.
        def _deco_init(func):
            return Pipeline(self, func, config)

        return _deco_init

    def consumer(self) -> Output:
        def _handler_wrap(func):
            return Output(app=self, handler=func)

        return _handler_wrap

    def standalone_consumer(self) -> StandAloneConsumer:
        def _handler_wrap(func):
            return StandAloneConsumer(app=self, handler=func)

        return _handler_wrap

    def consumer_iter(self) -> ConsumerIter:
        def _handler_wrap(func):
            return ConsumerIter(app=self, handler=func)

        return _handler_wrap

    def get_pipeline(self, name) -> Pipeline:
        return self.components.pipelines.get(name)

    def get_producer(self, name):
        return self.components.producers.get(name)

    def get_consumer(self, name):
        return self.components.consumers.get(name)



import click

from stairs import get_project

@click.group()
def general_cli():
    pass


@general_cli.command("admin")
def init_admin():
    from stairs.services.admin import server
    server.run_admin_server()


@general_cli.command("flushall")
def flushall():
    """
    Flush all queues which were used in the project.
    """
    p = get_project()

    msg = "Do you want to flush all jobs in your project?"
    if click.confirm(msg):
        for step in p.stepist_app.get_workers_steps():
            step.flush_all()



from stepist.flow import session
from stepist.flow.steps.next_step import call_next_step

from stairs.core.producer.utils import producer_retry
from stairs.core.app import components


class BatchProducer(components.AppProducer):
    DEFAULT_QUEUE_LIMIT = 10 ** 9

    def __init__(self, app, handler, simple_producer, queue_limit=None):
        self.app = app

        self.queue_limit = queue_limit or self.DEFAULT_QUEUE_LIMIT

        # The main generator which yields data
        self.handler = handler
        self.producer = simple_producer

        components.AppProducer.__init__(self, app)

    def __call__(self, *args, **kwargs):
        jobs_iterator = self.handler(*args, **kwargs)
        self.send_jobs(jobs_iterator)

    def flush(self):
        self.producer.stepist_step.flush_all()

    def send_jobs(self, job_iterator):
        with session.change_flow_ctx({}, {}):
            for job in job_iterator:
                # TODO: Make it more safe
                try:
                    self.send_job_to_producer(job)
                except Exception as e:
                    print("Something happened during batch generation, producer"
                          " was interrupted and jobs had forwarded to queue")
                    print("Run producer:flush for `%s` producer "
                          % self.producer.get_handler_name())
                    raise(e)

    @producer_retry(5, Exception)
    def send_job_to_producer(self, stepist_job):
        call_next_step(stepist_job, self.producer.stepist_step)

    def key(self):
        return self.handler.__name__



from stepist.flow.steps.step import StepData

import ujson
import pika


def get_connection(rmq_engine):
    impl = rmq_engine.pika_connection._impl
    return RMQConnection(impl.__class__, impl.params)


class RMQConnection:
    def __init__(self, _impl_class, params):
        self._impl_class = _impl_class
        self.params = params

        self.connection = None
        self.channel = None
        self.queues = None

    def init_connection(self):
        self.connection = pika.BlockingConnection(_impl_class=self._impl_class,
                                                  parameters=self.params)
        self.channel = self.connection.channel()

    def register_queue(self, step_key):
        self.channel.queue_declare(queue=step_key,
                                   auto_delete=False,
                                   passive=True)

    def add_job(self, step_key: str, data):
        if step_key not in self.queues:
            self.register_queue(step_key)

        step_data = StepData(data)
        step_data_str = ujson.dumps({'data': step_data.get_dict()})

        self.channel.basic_publish(
            exchange='',
            routing_key=step_key,
            body=step_data_str
        )




import ujson
from stepist.flow.steps.step import StepData

from redis import ConnectionPool, Redis


def get_connection(redis_engine):
    pool = redis_engine.redis_connection.connection_pool

    return RedisConnection(
        connection_class=pool.connection_class,
        connection_kwargs=pool.connection_kwargs,
        max_connections=pool.max_connections,
    )


class RedisConnection:
    def __init__(self, connection_class, connection_kwargs, max_connections):
        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.max_connections = max_connections

        self.connection = None
        self.channel = None
        self.queues = None

    def init_connection(self):
        self.connection = Redis(connection_pool=ConnectionPool(
            connection_class=self.connection_class,
            max_connections=self.max_connections,
            **self.connection_kwargs
        ))

    def add_job(self, step_key: str, data):
        step_data = StepData(data)
        step_data_str = ujson.dumps({'data': step_data.get_dict()})
        self.connection.lpush(self.redis_queue_key(step_key), step_data_str)

    def redis_queue_key(self, job_key):
        return "stepist::%s" % job_key




import click

from stairs.core.producer.producer import run_jobs_processor
from stairs.core.producer.producer import Producer
from stairs.core.producer.batch_producer import BatchProducer
from stairs.core.producer.spark_producer import SparkProducer
from stairs import get_project


@click.group()
def producer_cli():
    pass


@producer_cli.command("producer:run")
@click.argument('name')
@click.argument('pipelines', nargs=-1)
@click.option('--noprint', '-np', is_flag=True, help="Disable print", default=False)
@click.option('--nobatch_reading', '-nb', is_flag=True, help="Disable auto reading", default=False)
def run(name, pipelines, noprint, nobatch_reading):
    """
    Run your producer.

    In case you are running simple producer will generate jobs to pipelines and
    automatically exit.

    In case you are running batch producer will generate batches and automatically
    start batch reading process. To disable auto-reading specify --nobatch_reading
    flag.
    """
    project = get_project()
    project.set_verbose(not noprint)

    if project.verbose:
        print("Init producer session.")

    producer = get_project().get_producer_by_name(name)
    if producer is None:
        raise RuntimeError("Producer not found")

    if isinstance(producer, Producer):
        producer.run(pipelines or [], user_args=[], user_kwargs={})
        return

    if isinstance(producer, BatchProducer):
        producer(*[], **{})
        if project.verbose:
            print("Batch producer finished batch generation")
        if not nobatch_reading:
            if project.verbose:
                print("Starting batches reading process "
                      "and listening for a jobs ... ")
                print("Press Ctrl-C to cancel listening")
            batch_handler = producer.producer
            run_jobs_processor(get_project(),
                               [batch_handler],
                               custom_callbacks_keys=pipelines)

        return

    if isinstance(producer, SparkProducer):
        import time
        t1 = time.time()
        producer.run(pipelines or [], user_args=[], user_kwargs={})
        t2 = time.time()
        print(t2 - t1)
        return

    raise RuntimeError("Producer `%s` not found or not supported" % name)


@producer_cli.command("producer:run_jobs")
@click.argument("name", default=None)
@click.argument('pipelines', nargs=-1)
@click.option('--noprint', '-np', is_flag=True, help="Disable print")
def run_jobs(name, pipelines, noprint):
    """
    Run jobs which were generated by batch producer. You can apply this command
    both for simple producer and batch producer.
    """
    project = get_project()
    project.set_verbose(not noprint)

    if project.verbose:
        print("Starting batches reading process  and listening for a jobs ...")
        print("Press Ctrl-C to cancel listening")

    producers = []

    if name is None:
        for app in get_project().apps:
            producers = producers + app.components.producers.values()
    else:
        producer = get_project().get_producer_by_name(name)
        if producer is None:
            raise RuntimeError("Producer not found")

        if isinstance(producer, Producer):
           producers.append(producer)

        if isinstance(producer, BatchProducer):
            producers.append(producer.producer)

    if not producers:
        print("No producers found")

    run_jobs_processor(get_project(),
                       producers,
                       custom_callbacks_keys=pipelines)


@producer_cli.command("producer:flush_all")
@click.argument("name")
@click.option('--noconfirm', '-nс', is_flag=False,
              help="No confirmation message")
def flush_all(name, noconfirm):
    """
    Flush all jobs which were generated by specified producer.
    """
    if not noconfirm:
        msg = "Do you want to flush all jobs in %s producer?" % name
        if click.confirm(msg):
            noconfirm = True

    if noconfirm:
        producer = get_project().get_producer_by_name(name)
        if producer is None:
            raise RuntimeError("Producer not found")
        producer.flush()

    return


@producer_cli.command("producers")
def list_of_producers():
    """
    Returns producers list used in your project.
    """
    for app in get_project().apps:
        for producer_name in app.components.producers:
            print("%s.%s" % (app.app_name, producer_name))




import ujson

from stairs import get_project


def handle_sqs_event(pipeline, event):
    for record in event:
        handle_job(pipeline, **ujson.loads(record.body))


def handle_job(pipeline, **data):
    pipeline.step.receive_job(**data)
    get_project().run_pipelines(pipeline,
                                die_when_empty=True,
                                die_on_error=True)





class SignalsMixin:
    signals_on_app_created = []

    def on_app_created(self):
        def _wrap_handler(handler):
            self.signals_on_app_created.append(handler)

        return _wrap_handler


def send_signals(app, handlers):
    for handler in handlers:
        handler(app)



import stepist

from stepist.flow import session
from stepist.flow.steps.next_step import call_next_step

from stairs.core.producer.adapter import BaseProducerAdapter


class IterAdapter(BaseProducerAdapter):

    # db cursor
    handler = None

    app_input = None

    def __init__(self, app, handler, app_inputs, custom_inputs):
        self.app = app
        self.app_inputs = app_inputs or []
        self.custom_inputs = custom_inputs or dict()
        self.handler = handler

    def init_session(self):
        """
        Init generator session
        """
        pass

    def process(self, custom_pipelines_keys=None):
        try:
            for row in self.handler():
                with session.change_flow_ctx({}, {}):
                    if not isinstance(row, dict):
                        row = {'row': row}

                    pipelines_to_run = list(self.app_inputs)

                    for pipeline_key in custom_pipelines_keys or []:
                        if pipeline_key not in self.custom_inputs:
                            print("Pipeline %s not found" % pipeline_key)
                            print("Here available pipelines:")
                            self.print_custom_pipelines()
                        else:
                            pipelines_to_run.append(self.custom_inputs[pipeline_key])

                    if not pipelines_to_run:
                        print("No pipelines specified for producer")
                        print("If you have custom pipelines list them in command")

                    for input in pipelines_to_run:
                        call_next_step(row, input.step)

        except Exception:
            raise
    
    def flush_all(self):
        for input in self.app_inputs:
            input.step.flush_all()

    def print_custom_pipelines(self):
        for p in self.custom_inputs:
            print(p)



from stepist.flow import session
from stepist.flow.steps.next_step import call_next_step

from stairs.core.session import producer_session
from stairs.core.app import components
from stairs.core.producer.utils import producer_retry, custom_callbacks_to_dict


class Producer(components.AppProducer):
    """

    """
    DEFAULT_QUEUE_LIMIT = 10 ** 6

    def __init__(self, app, handler, default_callbacks: list,
                 custom_callbacks: list, queue_limit=None,
                 single_transaction=False):

        self.app = app

        self.queue_limit = queue_limit or self.DEFAULT_QUEUE_LIMIT
        self.single_transaction = single_transaction

        # The main generator which yields data
        self.handler = handler

        # Callbacks which should be run always
        self.default_callbacks = default_callbacks or []
        # Callbacks which should be run based on user console, input
        self.custom_callbacks = custom_callbacks_to_dict(custom_callbacks or [])

        # Stepist step basically to forward jobs to current producer
        # e.g. from Batch Producer
        self.stepist_step = self.app\
            .project\
            .stepist_app\
            .step(None,
                  as_worker=True,
                  unique_id=self.get_producer_id())(self.run_jobs)

        components.AppProducer.__init__(self, app)

    def __call__(self, *args, **kwargs):
        self.run(user_args=args, user_kwargs=kwargs)

    def run(self, custom_callbacks_keys: list = None,
            single_transaction: bool = False, user_args=None, user_kwargs=None):
        """
        Execute producer from console with specified args and kwargs.
        Also can have custom callbacks specified there.
        """
        custom_callbacks = []

        # Basic check for custom producers
        for custom_callback in custom_callbacks_keys or []:
            callback = self.custom_callbacks.get(custom_callback, None)
            if callback is None:
                print("Producer callback `%s` (another producer or pipeline"
                      ") not found." % custom_callback)
                exit()
            custom_callbacks.append(callback)

        # Basic check for callbacks
        if not custom_callbacks and not self.default_callbacks:
            print("No callbacks was found, specified default callback or use"
                  "custom callback")
            exit()

        callbacks_to_run = custom_callbacks + self.default_callbacks

        single_transaction = single_transaction or self.single_transaction
        user_args = user_args or []
        user_kwargs = user_kwargs or dict()

        # Running jobs from producer
        if not single_transaction:
            for job in self.handler(*user_args, **user_kwargs):
                self.send_job(job, callbacks_to_run)
        else:
            jobs_to_send = list(self.handler(*user_args, **user_kwargs))
            self.send_jobs(jobs_to_send, callbacks_to_run)

    def run_jobs(self, **kwargs):
        """
        Stepist Handler for executing forwarded jobs.

        Important(!) If you want to use custom callbacks, set them using:
        `producer_session.change_custom_callbacks` contextmanager
        """
        custom_callbacks = producer_session.get_custom_callbacks()
        self.run(custom_callbacks_keys=custom_callbacks,
                 user_kwargs=kwargs,
                 single_transaction=True)

    def send_job(self, job, callbacks_to_run):
        with session.change_flow_ctx({}, {}):

            # TODO: Make it more safe, it will good to run all callbacks
            # in one transaction, otherwise there is a chance to fail some
            # callback and duplication/lose data
            for callback in callbacks_to_run:
                self._job_to_stepist(job, callback.step)

    def send_jobs(self, jobs, callbacks_to_run):
        with session.change_flow_ctx({}, {}):

            # TODO: Make it more safe, it will good to run all callbacks
            # in one transaction, otherwise there is a chance to fail some
            # callback and duplication/lose data
            for callback in callbacks_to_run:
                self._job_to_stepist(jobs, callback.step, batch_data=True)

    @producer_retry(5, Exception)
    def _job_to_stepist(self, stepist_job, step, **kwargs):
        call_next_step(stepist_job, step, **kwargs)

    def get_producer_id(self):
        return "producer:%s:%s" % (self.app.app_name, self.handler.__name__)

    def get_handler_name(self):
        return self.handler.__name__

    def get_stepist_step(self):
        return self.stepist_step

    def flush(self):
        for pipeline in self.default_callbacks:
            pipeline.step.flush_all()

        for pipeline in self.custom_callbacks.values():
            pipeline.step.flush_all()

    def key(self):
        return self.get_handler_name()


def run_jobs_processor(project, producers_to_run, custom_callbacks_keys: list = None,
                       die_when_empty=False):
    """
    Executing forwarded jobs (from batch producer)
    """
    steps_to_run = [p.stepist_step for p in producers_to_run]
    with producer_session.change_custom_callbacks(custom_callbacks_keys):
        project \
           .stepist_app\
           .run(steps_to_run,
                die_on_error=True,
                die_when_empty=die_when_empty)